@@ -10,11 +10,13 @@
 #include "layout.h"
 #include "key.h"
 #include "clientlist.h"
+#include "monitor.h"
 
 #include <glib.h>
 #include <string.h>
 #include <stdlib.h>
 #include <stdio.h>
+#include <search.h>
 
 static char* completion_directions[]    = { "left", "right", "down", "up",NULL};
 static char* completion_focus_args[]    = { "-i", "-e", NULL };
@@ -38,7 +40,7 @@ static bool first_parameter_is_tag(int argc, char** argv, int pos);
 static bool first_parameter_is_flag(int argc, char** argv, int pos);
 static bool keybind_parameter_expected(int argc, char** argv, int pos);
 
-/* find out, if a parameter still expects a parameter at a certain index.
+/* find out, if a command still expects a parameter at a certain index.
  * only if this returns true, than a completion will be searched.
  *
  * if no match is found, then it defaults to "command still expects a
@@ -50,66 +52,80 @@ struct {
                         /* if current pos >= min_index */
     bool    (*function)(int argc, char** argv, int pos);
 } g_parameter_expected[] = {
-    { "quit",           0,  no_completion },
-    { "reload",         0,  no_completion },
-    { "version",        0,  no_completion },
-    { "list_monitors",  0,  no_completion },
-    { "list_commands",  0,  no_completion },
-    { "list_keybinds",  0,  no_completion },
-    { "add_monitor",    7,  no_completion },
-    { "bring",          2,  no_completion },
+    { "quit",           1,  no_completion },
+    { "reload",         1,  no_completion },
+    { "version",        1,  no_completion },
+    { "list_commands",  1,  no_completion },
+    { "list_monitors",  1,  no_completion },
+    { "list_keybinds",  1,  no_completion },
+    { "lock",           1,  no_completion },
+    { "unlock",         1,  no_completion },
+    { "keybind",        2,  keybind_parameter_expected },
+    { "keyunbind",      2,  no_completion },
+    { "mousebind",      3,  no_completion },
+    { "mouseunbind",    1,  no_completion },
     { "focus_nth",      2,  no_completion },
     { "cycle",          2,  no_completion },
     { "cycle_all",      3,  no_completion },
     { "cycle_layout",   LAYOUT_COUNT+2, no_completion },
-    { "cycle_monitor",  2,  no_completion },
-    { "close",          0,  no_completion },
-    { "close_or_remove",0,  no_completion },
-    { "dump",           2,  no_completion },
-    { "floating",       3,  no_completion },
-    { "floating",       2,  first_parameter_is_tag },
-    { "merge_tag",      3,  no_completion },
+    { "set_layout",     2,  no_completion },
+    { "close",          1,  no_completion },
+    { "close_or_remove",1,  no_completion },
+    { "split",          3,  no_completion },
     { "focus",          3,  no_completion },
     { "focus",          2,  first_parameter_is_flag },
-    { "shift",          3,  no_completion },
-    { "shift",          2,  first_parameter_is_flag },
-    { "split",          3,  no_completion },
-    { "fullscreen",     2,  no_completion },
-    { "pseudotile",     2,  no_completion },
-    { "pad",            6,  no_completion },
-    { "keybind",        2,  keybind_parameter_expected },
-    { "keyunbind",      2,  no_completion },
-    { "monitor_rect",   3,  no_completion },
-    { "mousebind",      3,  no_completion },
-    { "mouseunbind",    0,  no_completion },
-    { "move_monitor",   7,  no_completion },
-    { "layout",         2,  no_completion },
-    { "load",           3,  no_completion },
-    { "load",           2,  first_parameter_is_tag },
-    { "lock",           0,  no_completion },
-    { "unlock",         0,  no_completion },
-    { "move",           2,  no_completion },
-    { "move_index",     3,  no_completion },
     { "raise",          2,  no_completion },
     { "jumpto",         2,  no_completion },
-    { "rename",         3,  no_completion },
-    { "remove",         0,  no_completion },
-    { "remove_monitor", 2,  no_completion },
+    { "bring",          2,  no_completion },
     { "resize",         3,  no_completion },
-    { "unrule",         2,  no_completion },
-    { "use",            2,  no_completion },
-    { "use_index",      3,  no_completion },
-    { "add",            2,  no_completion },
+    { "focus_edge",     2,  no_completion },
+    { "shift_edge",     2,  no_completion },
+    { "shift",          3,  no_completion },
+    { "shift",          2,  first_parameter_is_flag },
+    { "remove",         1,  no_completion },
+    { "rotate",         1,  no_completion },
+    { "set",            3,  no_completion },
     { "get",            2,  no_completion },
     { "toggle",         2,  no_completion },
-    { "set",            3,  no_completion },
-    { "set_layout",     2,  no_completion },
+    { "cycle_monitor",  2,  no_completion },
+    { "focus_monitor",  2,  no_completion },
+    { "add",            2,  no_completion },
+    { "use",            2,  no_completion },
+    { "use_index",      3,  no_completion },
+    { "merge_tag",      3,  no_completion },
+    { "rename",         3,  no_completion },
+    { "move",           2,  no_completion },
+    { "move_index",     3,  no_completion },
+    { "lock_tag",       2,  no_completion },
+    { "unlock_tag",     2,  no_completion },
+    { "detect_monitors",1,  no_completion },
+    { "add_monitor",    7,  no_completion },
+    { "rename_monitor", 3,  no_completion },
+    { "remove_monitor", 2,  no_completion },
+    { "move_monitor",   7,  no_completion },
+    { "raise_monitor",  2,  no_completion },
+    { "stack",          2,  no_completion },
+    { "monitor_rect",   3,  no_completion },
+    { "pad",            6,  no_completion },
+    { "list_padding",   2,  no_completion },
+    { "layout",         3,  no_completion },
+    { "dump",           3,  no_completion },
+    { "load",           3,  no_completion },
+    { "load",           2,  first_parameter_is_tag },
     { "tag_status",     2,  no_completion },
+    { "floating",       3,  no_completion },
+    { "floating",       2,  first_parameter_is_tag },
+    { "unrule",         2,  no_completion },
+    { "fullscreen",     2,  no_completion },
+    { "pseudotile",     2,  no_completion },
+    { "getenv",         2,  no_completion },
+    { "setenv",         3,  no_completion },
+    { "unsetenv",       2,  no_completion },
     { 0 },
 };
 
 /* list of completions, if a line matches, then it will be used, the order
- * doesnot matter */
+ * does not matter */
 struct {
     char*   command;
     enum {
@@ -123,15 +139,17 @@ struct {
                         /* LE 3 matches position from 0 to 3 */
     /* === various methods, how to complete === */
     /* completion by function */
-    void (*function)(int argc, char** argv, int pos, GString** output);
+    void (*function)(int argc, char** argv, int pos, GString* output);
     /* completion by a list of strings */
     char** list;
 } g_completions[] = {
     /* name , relation, index,  completion method                   */
     { "add_monitor",    EQ, 2,  .function = complete_against_tags },
+    { "and",            GE, 1,  .function = complete_chain },
     { "bring",          EQ, 1,  .list = completion_special_winids },
     { "bring",          EQ, 1,  .function = complete_against_winids },
     { "cycle",          EQ, 1,  .list = completion_pm_one },
+    { "chain",          GE, 1,  .function = complete_chain },
     { "cycle_all",      EQ, 1,  .list = completion_cycle_all_args },
     { "cycle_all",      EQ, 1,  .list = completion_pm_one },
     { "cycle_all",      EQ, 2,  .list = completion_pm_one },
@@ -152,6 +170,8 @@ struct {
     { "merge_tag",      EQ, 2,  .function = complete_merge_tag },
     { "move",           EQ, 1,  .function = complete_against_tags },
     { "move_index",     EQ, 2,  .list = completion_use_index_args },
+    { "or",             GE, 1,  .function = complete_chain },
+    { "!",              GE, 1,  .function = complete_negate },
     { "pseudotile",     EQ, 1,  .list = completion_flag_args },
     { "keybind",        GE, 2,  .function = complete_against_keybind_command },
     { "keyunbind",      EQ, 1,  .list = completion_keyunbind_args },
@@ -162,6 +182,7 @@ struct {
     { "jumpto",         EQ, 1,  .list = completion_special_winids },
     { "jumpto",         EQ, 1,  .function = complete_against_winids },
     { "resize",         EQ, 1,  .list = completion_directions },
+    { "shift_edge",     EQ, 1,  .list = completion_directions },
     { "shift",          EQ, 1,  .list = completion_directions },
     { "shift",          EQ, 1,  .list = completion_focus_args },
     { "shift",          EQ, 2,  .list = completion_directions },
@@ -178,10 +199,21 @@ struct {
     { "use",            EQ, 1,  .function = complete_against_tags },
     { "use_index",      EQ, 1,  .list = completion_pm_one },
     { "use_index",      EQ, 2,  .list = completion_use_index_args },
+    { "focus_monitor",  EQ, 1,  .function = complete_against_monitors },
+    { "lock_tag",       EQ, 1,  .function = complete_against_monitors },
+    { "unlock_tag",     EQ, 1,  .function = complete_against_monitors },
+    { "remove_monitor", EQ, 1,  .function = complete_against_monitors },
+    { "move_monitor",   EQ, 1,  .function = complete_against_monitors },
+    { "raise_monitor",  EQ, 1,  .function = complete_against_monitors },
+    { "name_monitor",   EQ, 1,  .function = complete_against_monitors },
+    { "monitor_rect",   EQ, 1,  .function = complete_against_monitors },
+    { "pad",            EQ, 1,  .function = complete_against_monitors },
+    { "list_padding",   EQ, 1,  .function = complete_against_monitors },
+    { "tag_status",     EQ, 1,  .function = complete_against_monitors },
     { 0 },
 };
 
-int call_command(int argc, char** argv, GString** output) {
+int call_command(int argc, char** argv, GString* output) {
     if (argc <= 0) {
         return HERBST_COMMAND_NOT_FOUND;
     }
@@ -196,6 +228,8 @@ int call_command(int argc, char** argv, GString** output) {
         i++;
     }
     if (!bind) {
+        g_string_append_printf(output,
+            "error: Command \"%s\" not found\n", argv[0]);
         return HERBST_COMMAND_NOT_FOUND;
     }
     int status;
@@ -209,36 +243,35 @@ int call_command(int argc, char** argv, GString** output) {
 
 int call_command_no_output(int argc, char** argv) {
     GString* output = g_string_new("");
-    int status = call_command(argc, argv, &output);
+    int status = call_command(argc, argv, output);
     g_string_free(output, true);
     return status;
 }
 
-
-int list_commands(int argc, char** argv, GString** output)
+int list_commands(int argc, char** argv, GString* output)
 {
     int i = 0;
     while (g_commands[i].cmd.standard != NULL) {
-        *output = g_string_append(*output, g_commands[i].name);
-        *output = g_string_append(*output, "\n");
+        g_string_append(output, g_commands[i].name);
+        g_string_append(output, "\n");
         i++;
     }
     return 0;
 }
 
-void complete_against_list(char* needle, char** list, GString** output) {
+void complete_against_list(char* needle, char** list, GString* output) {
     size_t len = strlen(needle);
     while (*list) {
         char* name = *list;
         if (!strncmp(needle, name, len)) {
-            *output = g_string_append(*output, name);
-            *output = g_string_append(*output, "\n");
+            g_string_append(output, name);
+            g_string_append(output, "\n");
         }
         list++;
     }
 }
 
-void complete_against_tags(int argc, char** argv, int pos, GString** output) {
+void complete_against_tags(int argc, char** argv, int pos, GString* output) {
     char* needle;
     if (pos >= argc) {
         needle = "";
@@ -249,8 +282,38 @@ void complete_against_tags(int argc, char** argv, int pos, GString** output) {
     for (int i = 0; i < g_tags->len; i++) {
         char* name = g_array_index(g_tags, HSTag*, i)->name->str;
         if (!strncmp(needle, name, len)) {
-            *output = g_string_append(*output, name);
-            *output = g_string_append(*output, "\n");
+            g_string_append(output, name);
+            g_string_append(output, "\n");
+        }
+    }
+}
+
+void complete_negate(int argc, char** argv, int pos, GString* output) {
+    if (pos <= 0) {
+        return;
+    }
+    // Remove the ! from the argv
+    (void)SHIFT(argc, argv);
+    pos--;
+    // Complete as normal
+    complete_against_commands(argc, argv, pos, output);
+}
+
+void complete_against_monitors(int argc, char** argv, int pos, GString* output) {
+    char* needle;
+    if (pos >= argc) {
+        needle = "";
+    } else {
+        needle = argv[pos];
+    }
+    size_t len = strlen(needle);
+    for (int i = 0; i < monitor_count(); i++) {
+        GString* name = monitor_with_index(i)->name;
+        if (name != NULL) {
+            if (!strncmp(needle, name->str, len)) {
+                g_string_append(output, name->str);
+                g_string_append(output, "\n");
+            }
         }
     }
 }
@@ -258,20 +321,20 @@ void complete_against_tags(int argc, char** argv, int pos, GString** output) {
 struct wcd { /* window id completion data */
     char* needle;
     size_t needlelen;
-    GString** output;
+    GString* output;
 };
 
 static void add_winid_completion(void* key, HSClient* client, struct wcd* data)
 {
     char buf[100];
-    GString** out = data->output;
+    GString* out = data->output;
     snprintf(buf, LENGTH(buf), "0x%lx\n", client->window);
     if (!strncmp(buf, data->needle, data->needlelen)) {
-        *out = g_string_append(*out, buf);
+        g_string_append(out, buf);
     }
 }
 
-void complete_against_winids(int argc, char** argv, int pos, GString** output) {
+void complete_against_winids(int argc, char** argv, int pos, GString* output) {
     struct wcd data;
     if (pos >= argc) {
         data.needle = "";
@@ -283,9 +346,8 @@ void complete_against_winids(int argc, char** argv, int pos, GString** output) {
     clientlist_foreach((GHFunc)add_winid_completion, &data);
 }
 
-
-void complete_merge_tag(int argc, char** argv, int pos, GString** output) {
-    char* first = (argc >= 1) ? argv[1] : "";
+void complete_merge_tag(int argc, char** argv, int pos, GString* output) {
+    char* first = (argc > 1) ? argv[1] : "";
     char* needle;
     if (pos >= argc) {
         needle = "";
@@ -300,13 +362,13 @@ void complete_merge_tag(int argc, char** argv, int pos, GString** output) {
             continue;
         }
         if (!strncmp(needle, name, len)) {
-            *output = g_string_append(*output, name);
-            *output = g_string_append(*output, "\n");
+            g_string_append(output, name);
+            g_string_append(output, "\n");
         }
     }
 }
 
-void complete_against_settings(int argc, char** argv, int pos, GString** output)
+void complete_against_settings(int argc, char** argv, int pos, GString* output)
 {
     char* needle;
     if (pos >= argc) {
@@ -323,13 +385,13 @@ void complete_against_settings(int argc, char** argv, int pos, GString** output)
         }
         // only check the first len bytes
         if (!strncmp(needle, g_settings[i].name, len)) {
-            *output = g_string_append(*output, g_settings[i].name);
-            *output = g_string_append(*output, "\n");
+            g_string_append(output, g_settings[i].name);
+            g_string_append(output, "\n");
         }
     }
 }
 
-void complete_against_keybinds(int argc, char** argv, int pos, GString** output) {
+void complete_against_keybinds(int argc, char** argv, int pos, GString* output) {
     char* needle;
     if (pos >= argc) {
         needle = "";
@@ -356,12 +418,12 @@ bool parameter_expected(int argc, char** argv, int pos) {
     return true;
 }
 
-int complete_command(int argc, char** argv, GString** output) {
+int complete_command(int argc, char** argv, GString* output) {
     // usage: complete POSITION command to complete ...
     if (argc < 2) {
-        return HERBST_INVALID_ARGUMENT;
+        return HERBST_NEED_MORE_ARGS;
     }
-    // index must be between first and als arg of "commmand to complete ..."
+    // index must be between first and last arg of "command to complete ..."
     int position = CLAMP(atoi(argv[1]), 0, argc-2);
     (void)SHIFT(argc, argv);
     (void)SHIFT(argc, argv);
@@ -369,7 +431,7 @@ int complete_command(int argc, char** argv, GString** output) {
 }
 
 void complete_against_keybind_command(int argc, char** argv, int position,
-                                      GString** output) {
+                                      GString* output) {
     if (argc <  2 || position < 2) {
         return;
     }
@@ -377,7 +439,7 @@ void complete_against_keybind_command(int argc, char** argv, int position,
 }
 
 int complete_against_commands(int argc, char** argv, int position,
-                              GString** output) {
+                              GString* output) {
     // complete command
     if (position == 0) {
         char* str = (argc >= 1) ? argv[0] : "";
@@ -386,8 +448,8 @@ int complete_against_commands(int argc, char** argv, int position,
         while (g_commands[i].cmd.standard != NULL) {
             // only check the first len bytes
             if (!strncmp(str, g_commands[i].name, len)) {
-                *output = g_string_append(*output, g_commands[i].name);
-                *output = g_string_append(*output, "\n");
+                g_string_append(output, g_commands[i].name);
+                g_string_append(output, "\n");
             }
             i++;
         }
@@ -428,6 +490,61 @@ int complete_against_commands(int argc, char** argv, int position,
     return 0;
 }
 
+static int strpcmp(const void* key, const void* val) {
+    return strcmp((const char*) key, *(const char**)val);
+}
+
+static void complete_chain_helper(int argc, char** argv, int position,
+                                  GString* output) {
+    /* argv entries:
+     * argv[0]      == the command separator
+     * argv[1]      == an arbitrary command name
+     * argv[2..]    == its arguments or a separator
+     */
+    if (position <= 0 || argc <= 1) {
+        return;
+    }
+    char* separator = argv[0];
+    (void)SHIFT(argc, argv);
+    position--;
+
+    /* find the next separator */
+    size_t uargc = argc;
+    char** next_sep = lfind(separator, argv, &uargc, sizeof(*argv), strpcmp);
+    int next_sep_idx = next_sep - argv;
+
+    if (!next_sep || next_sep_idx >= position) {
+        /* try to complete at the desired position */
+        char* needle = (position < argc) ? argv[position] : "";
+        complete_against_commands(argc, argv, position, output);
+        /* at least the command name is required
+         * so don't complete at position 0 */
+        if (position != 0 && 0 == strncmp(needle, separator, strlen(needle))) {
+            g_string_append_printf(output, "%s\n", separator);
+        }
+    } else {
+        /* remove arguments so that the next separator becomes argv[0] */
+        position -= next_sep_idx;
+        argc     -= next_sep_idx;
+        argv     += next_sep_idx;
+        complete_chain_helper(argc, argv, position, output);
+    }
+}
+
+void complete_chain(int argc, char** argv, int position, GString* output) {
+    if (position <= 1) {
+        return;
+    }
+    /* remove the chain command name "chain" from the argv */
+    (void)SHIFT(argc, argv);
+    position--;
+
+    /* do the actual work in the helper that always expects
+     * {separator, firstcommand, ...}
+     */
+    complete_chain_helper(argc, argv, position, output);
+}
+
 bool first_parameter_is_tag(int argc, char** argv, int pos) {
     // only complete if first parameter is a valid tag
     if (argc >= 2 && find_tag(argv[1]) && pos == 2) {
@@ -457,3 +574,61 @@ bool keybind_parameter_expected(int argc, char** argv, int pos) {
     return parameter_expected(argc - 2, argv + 2, pos - 2);
 }
 
+int command_chain(char* separator, bool (*condition)(int laststatus),
+                  int argc, char** argv, GString* output) {
+    size_t uargc = argc;
+    printf("finding %s in %d elements\n", separator, argc);
+    char** next_sep = lfind(separator, argv, &uargc, sizeof(*argv), strpcmp);
+    int command_argc = next_sep ? (int)(next_sep - argv) : argc;
+    int status = call_command(command_argc, argv, output);
+    if (condition && false == condition(status)) {
+        return status;
+    }
+    argc -= command_argc + 1;
+    argv += command_argc + 1;
+    if (argc <= 0) {
+        return status;
+    }
+    return command_chain(separator, condition, argc, argv, output);
+}
+
+static bool int_is_zero(int x) {
+    return x == 0;
+}
+
+static bool int_is_not_zero(int x) {
+    return x != 0;
+}
+
+typedef struct {
+    char* cmd;
+    bool (*condition)(int);
+} Cmd2Condition;
+
+static Cmd2Condition g_cmd2condition[] = {
+    { "and",    int_is_zero         },
+    { "or",     int_is_not_zero     },
+};
+
+int command_chain_command(int argc, char** argv, GString* output) {
+    Cmd2Condition* cmd;
+    cmd = STATIC_TABLE_FIND_STR(Cmd2Condition, g_cmd2condition, cmd, argv[0]);
+    (void)SHIFT(argc, argv);
+    if (argc <= 1) {
+        return HERBST_NEED_MORE_ARGS;
+    }
+    char* separator = argv[0];
+    (void)SHIFT(argc, argv);
+    bool (*condition)(int) = cmd ? cmd->condition : NULL;
+    return command_chain(separator, condition, argc, argv, output);
+}
+
+int negate_command(int argc, char** argv, GString* output) {
+    if (argc <= 1) {
+        return HERBST_NEED_MORE_ARGS;
+    }
+    (void)SHIFT(argc, argv);
+    int status = call_command(argc, argv, output);
+    return (!status);
+}
+
