@@ -29,6 +29,7 @@
 #include <sys/select.h>
 #include <sys/wait.h>
 #include <assert.h>
+#include <errno.h>
 // gui
 #include <X11/Xlib.h>
 #include <X11/Xproto.h>
@@ -49,17 +50,20 @@ typedef void (*HandlerTable[LASTEvent]) (XEvent*);
 
 int quit();
 int reload();
-int version(int argc, char* argv[], GString** result);
-int print_layout_command(int argc, char** argv, GString** result);
-int load_command(int argc, char** argv, GString** result);
-int print_tag_status_command(int argc, char** argv, GString** result);
+int version(int argc, char* argv[], GString* output);
+int print_layout_command(int argc, char** argv, GString* output);
+int load_command(int argc, char** argv, GString* output);
+int print_tag_status_command(int argc, char** argv, GString* output);
 void execute_autostart_file();
-int raise_command(int argc, char** argv);
+int raise_command(int argc, char** argv, GString* output);
 int spawn(int argc, char** argv);
 int wmexec(int argc, char** argv);
 static void remove_zombies(int signal);
 int custom_hook_emit(int argc, char** argv);
-int jumpto_command(int argc, char** argv);
+int jumpto_command(int argc, char** argv, GString* output);
+int getenv_command(int argc, char** argv, GString* output);
+int setenv_command(int argc, char** argv, GString* output);
+int unsetenv_command(int argc, char** argv, GString* output);
 
 // handler for X-Events
 void buttonpress(XEvent* event);
@@ -88,53 +92,58 @@ CommandBinding g_commands[] = {
     CMD_BIND(             "set_monitors",   set_monitor_rects_command),
     CMD_BIND(             "disjoin_rects",  disjoin_rects_command),
     CMD_BIND(             "list_keybinds",  key_list_binds),
-    CMD_BIND_NO_OUTPUT(   "keybind",        keybind),
-    CMD_BIND_NO_OUTPUT(   "keyunbind",      keyunbind),
-    CMD_BIND_NO_OUTPUT(   "mousebind",      mouse_bind_command),
+    CMD_BIND(             "list_padding",   list_padding),
+    CMD_BIND(             "keybind",        keybind),
+    CMD_BIND(             "keyunbind",      keyunbind),
+    CMD_BIND(             "mousebind",      mouse_bind_command),
     CMD_BIND_NO_OUTPUT(   "mouseunbind",    mouse_unbind_all),
     CMD_BIND_NO_OUTPUT(   "spawn",          spawn),
     CMD_BIND_NO_OUTPUT(   "wmexec",         wmexec),
     CMD_BIND_NO_OUTPUT(   "emit_hook",      custom_hook_emit),
-    CMD_BIND_NO_OUTPUT(   "bring",          frame_current_bring),
+    CMD_BIND(             "bring",          frame_current_bring),
     CMD_BIND_NO_OUTPUT(   "focus_nth",      frame_current_set_selection),
     CMD_BIND_NO_OUTPUT(   "cycle",          frame_current_cycle_selection),
     CMD_BIND_NO_OUTPUT(   "cycle_all",      cycle_all_command),
-    CMD_BIND_NO_OUTPUT(   "cycle_layout",   frame_current_cycle_client_layout),
+    CMD_BIND(             "cycle_layout",   frame_current_cycle_client_layout),
     CMD_BIND_NO_OUTPUT(   "close",          window_close_current),
     CMD_BIND_NO_OUTPUT(   "close_or_remove",close_or_remove_command),
-    CMD_BIND_NO_OUTPUT(   "split",          frame_split_command),
-    CMD_BIND_NO_OUTPUT(   "resize",         frame_change_fraction_command),
-    CMD_BIND_NO_OUTPUT(   "focus",          frame_focus_command),
-    CMD_BIND_NO_OUTPUT(   "shift",          frame_move_window_command),
+    CMD_BIND(             "split",          frame_split_command),
+    CMD_BIND(             "resize",         frame_change_fraction_command),
+    CMD_BIND(             "focus_edge",     frame_focus_edge),
+    CMD_BIND(             "focus",          frame_focus_command),
+    CMD_BIND(             "shift_edge",     frame_move_window_edge),
+    CMD_BIND(             "shift",          frame_move_window_command),
+    CMD_BIND(             "shift_to_monitor",shift_to_monitor),
     CMD_BIND_NO_OUTPUT(   "remove",         frame_remove_command),
-    CMD_BIND_NO_OUTPUT(   "set",            settings_set_command),
-    CMD_BIND_NO_OUTPUT(   "toggle",         settings_toggle),
-    CMD_BIND_NO_OUTPUT(   "cycle_value",    settings_cycle_value),
+    CMD_BIND(             "set",            settings_set_command),
+    CMD_BIND(             "toggle",         settings_toggle),
+    CMD_BIND(             "cycle_value",    settings_cycle_value),
     CMD_BIND_NO_OUTPUT(   "cycle_monitor",  monitor_cycle_command),
-    CMD_BIND_NO_OUTPUT(   "focus_monitor",  monitor_focus_command),
+    CMD_BIND(             "focus_monitor",  monitor_focus_command),
     CMD_BIND(             "get",            settings_get),
-    CMD_BIND_NO_OUTPUT(   "add",            tag_add_command),
-    CMD_BIND_NO_OUTPUT(   "use",            monitor_set_tag_command),
-    CMD_BIND_NO_OUTPUT(   "use_index",      monitor_set_tag_by_index_command),
-    CMD_BIND_NO_OUTPUT(   "jumpto",         jumpto_command),
+    CMD_BIND(             "add",            tag_add_command),
+    CMD_BIND(             "use",            monitor_set_tag_command),
+    CMD_BIND(             "use_index",      monitor_set_tag_by_index_command),
+    CMD_BIND(             "jumpto",         jumpto_command),
     CMD_BIND(             "floating",       tag_set_floating_command),
     CMD_BIND_NO_OUTPUT(   "fullscreen",     client_set_property_command),
     CMD_BIND_NO_OUTPUT(   "pseudotile",     client_set_property_command),
     CMD_BIND(             "tag_status",     print_tag_status_command),
-    CMD_BIND_NO_OUTPUT(   "merge_tag",      tag_remove_command),
-    CMD_BIND_NO_OUTPUT(   "rename",         tag_rename_command),
-    CMD_BIND_NO_OUTPUT(   "move",           tag_move_window_command),
+    CMD_BIND(             "merge_tag",      tag_remove_command),
+    CMD_BIND(             "rename",         tag_rename_command),
+    CMD_BIND(             "move",           tag_move_window_command),
     CMD_BIND_NO_OUTPUT(   "rotate",         layout_rotate_command),
-    CMD_BIND_NO_OUTPUT(   "move_index",     tag_move_window_by_index_command),
-    CMD_BIND_NO_OUTPUT(   "add_monitor",    add_monitor_command),
-    CMD_BIND_NO_OUTPUT(   "raise_monitor",  monitor_raise_command),
-    CMD_BIND_NO_OUTPUT(   "remove_monitor", remove_monitor_command),
-    CMD_BIND_NO_OUTPUT(   "move_monitor",   move_monitor_command),
+    CMD_BIND(             "move_index",     tag_move_window_by_index_command),
+    CMD_BIND(             "add_monitor",    add_monitor_command),
+    CMD_BIND(             "raise_monitor",  monitor_raise_command),
+    CMD_BIND(             "remove_monitor", remove_monitor_command),
+    CMD_BIND(             "move_monitor",   move_monitor_command),
+    CMD_BIND(             "rename_monitor", rename_monitor_command),
     CMD_BIND(             "monitor_rect",   monitor_rect_command),
-    CMD_BIND_NO_OUTPUT(   "pad",            monitor_set_pad_command),
-    CMD_BIND_NO_OUTPUT(   "raise",          raise_command),
-    CMD_BIND_NO_OUTPUT(   "rule",           rule_add_command),
-    CMD_BIND_NO_OUTPUT(   "unrule",         rule_remove_command),
+    CMD_BIND(             "pad",            monitor_set_pad_command),
+    CMD_BIND(             "raise",          raise_command),
+    CMD_BIND(             "rule",           rule_add_command),
+    CMD_BIND(             "unrule",         rule_remove_command),
     CMD_BIND(             "layout",         print_layout_command),
     CMD_BIND(             "stack",          print_stack_command),
     CMD_BIND(             "dump",           print_layout_command),
@@ -142,12 +151,21 @@ CommandBinding g_commands[] = {
     CMD_BIND(             "complete",       complete_command),
     CMD_BIND_NO_OUTPUT(   "lock",           monitors_lock_command),
     CMD_BIND_NO_OUTPUT(   "unlock",         monitors_unlock_command),
-    CMD_BIND_NO_OUTPUT(   "set_layout",     frame_current_set_client_layout),
-    CMD_BIND_NO_OUTPUT(   "detect_monitors",detect_monitors_command),
+    CMD_BIND(             "lock_tag",       monitor_lock_tag_command),
+    CMD_BIND(             "unlock_tag",     monitor_unlock_tag_command),
+    CMD_BIND(             "set_layout",     frame_current_set_client_layout),
+    CMD_BIND(             "detect_monitors",detect_monitors_command),
+    CMD_BIND(             "chain",          command_chain_command),
+    CMD_BIND(             "and",            command_chain_command),
+    CMD_BIND(             "or",             command_chain_command),
+    CMD_BIND(             "!",              negate_command),
+    CMD_BIND(             "getenv",         getenv_command),
+    CMD_BIND(             "setenv",         setenv_command),
+    CMD_BIND(             "unsetenv",       unsetenv_command),
     {{ NULL }}
 };
 
-// core funcitons
+// core functions
 int quit() {
     g_aboutToQuit = true;
     return 0;
@@ -159,52 +177,64 @@ int reload() {
     return 0;
 }
 
-int version(int argc, char* argv[], GString** result) {
+int version(int argc, char* argv[], GString* output) {
     (void) argc;
     (void) argv;
-    *result = g_string_assign(*result, HERBSTLUFT_VERSION_STRING);
+    g_string_append(output, HERBSTLUFT_VERSION_STRING);
     return 0;
 }
 
 // prints or dumps the layout of an given tag
-// first argument tells wether to print or to dump
-int print_layout_command(int argc, char** argv, GString** result) {
+// first argument tells whether to print or to dump
+int print_layout_command(int argc, char** argv, GString* output) {
     HSTag* tag = NULL;
-    if (argc >= 2) {
+    // an empty argv[1] means current focused tag
+    if (argc >= 2 && argv[1][0] != '\0') {
         tag = find_tag(argv[1]);
-    }
-    // if no tag was found
-    if (!tag) {
+        if (!tag) {
+            g_string_append_printf(output,
+                "%s: Tag \"%s\" not found\n", argv[0], argv[1]);
+            return HERBST_INVALID_ARGUMENT;
+        }
+    } else { // use current tag
         HSMonitor* m = get_current_monitor();
         tag = m->tag;
     }
     assert(tag != NULL);
+
+    HSFrame* frame = lookup_frame(tag->frame, argc >= 3 ? argv[2] : "");
     if (argc > 0 && !strcmp(argv[0], "dump")) {
-        dump_frame_tree(tag->frame, result);
+        dump_frame_tree(frame, output);
     } else {
-        print_tag_tree(tag, result);
+        print_frame_tree(frame, output);
     }
     return 0;
 }
 
-int load_command(int argc, char** argv, GString** result) {
+int load_command(int argc, char** argv, GString* output) {
     // usage: load TAG LAYOUT
     HSTag* tag = NULL;
     if (argc < 2) {
-        return HERBST_INVALID_ARGUMENT;
+        return HERBST_NEED_MORE_ARGS;
     }
     char* layout_string = argv[1];
     if (argc >= 3) {
         tag = find_tag(argv[1]);
         layout_string = argv[2];
-    }
-    // if no tag was found
-    if (!tag) {
+        if (!tag) {
+            g_string_append_printf(output,
+                "%s: Tag \"%s\" not found\n", argv[0], argv[1]);
+            return HERBST_INVALID_ARGUMENT;
+        }
+    } else { // use current tag
         HSMonitor* m = get_current_monitor();
         tag = m->tag;
     }
     assert(tag != NULL);
-    char* rest = load_frame_tree(tag->frame, layout_string, result);
+    char* rest = load_frame_tree(tag->frame, layout_string, output);
+    if (output->len > 0) {
+        g_string_prepend(output, "load: ");
+    }
     tag_set_flags_dirty(); // we probably changed some window positions
     // arrange monitor
     HSMonitor* m = find_monitor_with_tag(tag);
@@ -218,27 +248,35 @@ int load_command(int argc, char** argv, GString** result) {
         frame_hide_recursive(tag->frame);
     }
     if (!rest) {
+        g_string_append_printf(output,
+            "%s: Error while parsing!\n", argv[0]);
         return HERBST_INVALID_ARGUMENT;
     }
-    if (rest[0] != '\0') { // if string wasnot parsed completely
-        g_string_append_printf(*result,
-            "%s: layout description was too long\n", argv[0]);
-        g_string_append_printf(*result,
+    if (rest[0] != '\0') { // if string was not parsed completely
+        g_string_append_printf(output,
+            "%s: Layout description was too long\n", argv[0]);
+        g_string_append_printf(output,
             "%s: \"%s\" has not been parsed\n", argv[0], rest);
         return HERBST_INVALID_ARGUMENT;
     }
     return 0;
 }
 
-int print_tag_status_command(int argc, char** argv, GString** result) {
+int print_tag_status_command(int argc, char** argv, GString* output) {
+    HSMonitor* monitor;
     int monitor_index = g_cur_monitor;
     if (argc >= 2) {
-        monitor_index = atoi(argv[1]);
+        monitor = string_to_monitor(argv[1]);
+    } else {
+        monitor = get_current_monitor();
+    }
+    if (monitor == NULL) {
+        g_string_append_printf(output,
+            "%s: Monitor \"%s\" not found!\n", argv[0], argv[1]);
+        return HERBST_INVALID_ARGUMENT;
     }
-    monitor_index = CLAMP(monitor_index, 0, monitor_count());
     tag_update_flags();
-    HSMonitor* monitor = monitor_with_index(monitor_index);
-    *result = g_string_append_c(*result, '\t');
+    g_string_append_c(output, '\t');
     for (int i = 0; i < g_tags->len; i++) {
         HSTag* tag = g_array_index(g_tags, HSTag*, i);
         // print flags
@@ -261,9 +299,9 @@ int print_tag_status_command(int argc, char** argv, GString** result) {
         if (tag->flags & TAG_FLAG_URGENT) {
             c = '!';
         }
-        *result = g_string_append_c(*result, c);
-        *result = g_string_append(*result, tag->name->str);
-        *result = g_string_append_c(*result, '\t');
+        g_string_append_c(output, c);
+        g_string_append(output, tag->name->str);
+        g_string_append_c(output, '\t');
     }
     return 0;
 }
@@ -276,8 +314,7 @@ int custom_hook_emit(int argc, char** argv) {
 // spawn() heavily inspired by dwm.c
 int spawn(int argc, char** argv) {
     if (argc < 2) {
-        fprintf(stderr, "spawn: too few parameters\n");
-        return HERBST_INVALID_ARGUMENT;
+        return HERBST_NEED_MORE_ARGS;
     }
     if (fork() == 0) {
         // only look in child
@@ -325,11 +362,11 @@ int wmexec(int argc, char** argv) {
     return EXIT_SUCCESS;
 }
 
-int raise_command(int argc, char** argv) {
+int raise_command(int argc, char** argv, GString* output) {
     Window win;
     HSClient* client = NULL;
     win = string_to_client((argc > 1) ? argv[1] : "", &client);
-if (client) {
+    if (client) {
         client_raise(client);
     } else {
         XRaiseWindow(g_display, win);
@@ -337,15 +374,62 @@ if (client) {
     return 0;
 }
 
-int jumpto_command(int argc, char** argv) {
+int jumpto_command(int argc, char** argv, GString* output) {
+    if (argc < 2) {
+        return HERBST_NEED_MORE_ARGS;
+    }
     HSClient* client = NULL;
-    string_to_client((argc > 1) ? argv[1] : "", &client);
+    string_to_client(argv[1], &client);
     if (client) {
         focus_window(client->window, true, true);
+        return 0;
+    } else {
+        g_string_append_printf(output,
+            "%s: Could not find client", argv[0]);
+        if (argc > 1) {
+            g_string_append_printf(output, " \"%s\".\n", argv[1]);
+        } else {
+            g_string_append(output, ".\n");
+        }
+        return HERBST_INVALID_ARGUMENT;
+    }
+}
+
+int getenv_command(int argc, char** argv, GString* output) {
+    if (argc < 2) {
+        return HERBST_NEED_MORE_ARGS;
+    }
+    char* envvar = getenv(argv[1]);
+    if (envvar == NULL) {
+        return HERBST_ENV_UNSET;
+    }
+    g_string_append_printf(output, "%s\n", envvar);
+    return 0;
+}
+
+int setenv_command(int argc, char** argv, GString* output) {
+    if (argc < 3) {
+        return HERBST_NEED_MORE_ARGS;
+    }
+    if (setenv(argv[1], argv[2], 1) != 0) {
+        g_string_append_printf(output,
+            "%s: Could not set environment variable: %s\n", argv[0], strerror(errno));
+        return HERBST_UNKNOWN_ERROR;
     }
     return 0;
 }
 
+int unsetenv_command(int argc, char** argv, GString* output) {
+    if (argc < 2) {
+        return HERBST_NEED_MORE_ARGS;
+    }
+    if (unsetenv(argv[1]) != 0) {
+        g_string_append_printf(output,
+            "%s: Could not unset environment variable: %s\n", argv[0], strerror(errno));
+        return HERBST_UNKNOWN_ERROR;
+    }
+    return 0;
+}
 
 // handle x-events:
 
@@ -391,7 +475,6 @@ void event_on_configure(XEvent event) {
     }
 }
 
-
 // from dwm.c
 /* There's no way to check accesses to destroyed windows, thus those cases are
  * ignored (especially on UnmapNotify's).  Other types of errors call Xlibs
@@ -419,7 +502,6 @@ int xerror(Display *dpy, XErrorEvent *ee) {
     return g_xerrorxlib(dpy, ee); /* may call exit */
 }
 
-
 int xerrordummy(Display *dpy, XErrorEvent *ee) {
     return 0;
 }
@@ -432,7 +514,6 @@ int xerrorstart(Display *dpy, XErrorEvent *ee) {
     return -1;
 }
 
-
 // from dwm.c
 void checkotherwm(void) {
     g_otherwm = False;
@@ -446,7 +527,6 @@ void checkotherwm(void) {
     XSync(g_display, False);
 }
 
-
 // scan for windows and add them to the list of managed clients
 // from dwm.c
 void scan(void) {
@@ -489,16 +569,16 @@ void execute_autostart_file() {
         } else {
             char* home = getenv("HOME");
             if (!home) {
-                g_warning("Willnot run autostart file. "
+                g_warning("Will not run autostart file. "
                           "Neither $HOME or $XDG_CONFIG_HOME is set.\n");
                 return;
             }
             path = g_string_new(home);
-            path = g_string_append_c(path, G_DIR_SEPARATOR);
-            path = g_string_append(path, ".config");
+            g_string_append_c(path, G_DIR_SEPARATOR);
+            g_string_append(path, ".config");
         }
-        path = g_string_append_c(path, G_DIR_SEPARATOR);
-        path = g_string_append(path, HERBSTLUFT_AUTOSTART);
+        g_string_append_c(path, G_DIR_SEPARATOR);
+        g_string_append(path, HERBSTLUFT_AUTOSTART);
     }
     if (0 == fork()) {
         if (g_display) {
@@ -548,7 +628,6 @@ static void parse_arguments(int argc, char** argv) {
                 g_initial_monitors_locked = 1;
                 break;
             default:
-                fprintf(stderr, "unknown option `%s'\n", argv[optind]);
                 exit(EXIT_FAILURE);
         }
     }
@@ -649,22 +728,26 @@ void createnotify(XEvent* event) {
         ipc_add_connection(event->xcreatewindow.window);
     }
 }
+
 void configurerequest(XEvent* event) {
     HSDebug("name is: ConfigureRequest\n");
     event_on_configure(*event);
 }
+
 void configurenotify(XEvent* event) {
     if (event->xconfigure.window == g_root &&
         settings_find("auto_detect_monitors")->value.i) {
         char* args[] = { "detect_monitors" };
-        detect_monitors_command(LENGTH(args), args);
+        detect_monitors_command(LENGTH(args), args, NULL);
     }
     // HSDebug("name is: ConfigureNotify\n");
 }
+
 void destroynotify(XEvent* event) {
     // try to unmanage it
     unmanage_client(event->xdestroywindow.window);
 }
+
 void enternotify(XEvent* event) {
     XCrossingEvent *ce = &event->xcrossing;
     HSDebug("name is: EnterNotify, focus = %d\n", event->xcrossing.focus);
@@ -675,16 +758,20 @@ void enternotify(XEvent* event) {
         focus_window(ce->window, false, true);
     }
 }
+
 void expose(XEvent* event) {
     HSDebug("name is: Expose\n");
 }
+
 void focusin(XEvent* event) {
     HSDebug("name is: FocusIn\n");
 }
+
 void keypress(XEvent* event) {
     HSDebug("name is: KeyPress\n");
     handle_key_press(event);
 }
+
 void mappingnotify(XEvent* event) {
     {
         // regrab when keyboard map changes
@@ -696,9 +783,11 @@ void mappingnotify(XEvent* event) {
         }
     }
 }
+
 void motionnotify(XEvent* event) {
     handle_motion_event(event);
 }
+
 void mapnotify(XEvent* event) {
     HSDebug("name is: MapNotify\n");
     HSClient* c;
@@ -710,6 +799,7 @@ void mapnotify(XEvent* event) {
         client_update_title(c);
     }
 }
+
 void maprequest(XEvent* event) {
     HSDebug("name is: MapRequest\n");
     XMapRequestEvent* mapreq = &event->xmaprequest;
@@ -733,12 +823,12 @@ void maprequest(XEvent* event) {
 }
 
 void propertynotify(XEvent* event) {
-    // printf("name is: PropertyNotify\n"); 
+    // printf("name is: PropertyNotify\n");
     XPropertyEvent *ev = &event->xproperty;
     HSClient* client;
     if (ev->state == PropertyNewValue) {
         if (is_ipc_connectable(event->xproperty.window)) {
-            ipc_handle_connection(event->xproperty.window, false);
+            ipc_handle_connection(event->xproperty.window);
         } else if((client = get_client_from_window(ev->window))) {
             switch (ev->atom) {
                 case XA_WM_HINTS:
@@ -753,12 +843,12 @@ void propertynotify(XEvent* event) {
         }
     }
 }
+
 void unmapnotify(XEvent* event) {
     HSDebug("name is: UnmapNotify for %lx\n", event->xunmap.window);
     unmanage_client(event->xunmap.window);
 }
 
-
 /* ---- */
 /* main */
 /* ---- */
@@ -821,7 +911,7 @@ int main(int argc, char* argv[]) {
         g_modules[i].destroy();
     }
     XCloseDisplay(g_display);
-    // check if wie shall restart an other window manager
+    // check if we shall restart an other window manager
     if (g_exec_before_quit) {
         if (g_exec_args) {
             // do actual exec
@@ -840,5 +930,3 @@ int main(int argc, char* argv[]) {
     return EXIT_SUCCESS;
 }
 
-
-
