@@ -2,9 +2,12 @@
  *
  * This software is licensed under the "Simplified BSD License".
  * See LICENSE for details */
+
 #include <assert.h>
 #include <string.h>
 #include <stdio.h>
+#include <stdbool.h>
+#include <ctype.h>
 #ifdef XINERAMA
 #include <X11/extensions/Xinerama.h>
 #endif /* XINERAMA */
@@ -25,6 +28,7 @@
 int* g_monitors_locked;
 int* g_swap_monitors_to_get_tag;
 int* g_smart_frame_surroundings;
+int* g_mouse_recenter_gap;
 HSStack* g_monitor_stack;
 GArray*     g_monitors; // Array of HSMonitor*
 
@@ -41,6 +45,7 @@ void monitor_init() {
     g_monitors = g_array_new(false, false, sizeof(HSMonitor*));
     g_swap_monitors_to_get_tag = &(settings_find("swap_monitors_to_get_tag")->value.i);
     g_smart_frame_surroundings = &(settings_find("smart_frame_surroundings")->value.i);
+    g_mouse_recenter_gap       = &(settings_find("mouse_recenter_gap")->value.i);
     g_monitor_stack = stack_create();
 }
 
@@ -55,7 +60,6 @@ void monitor_destroy() {
     g_array_free(g_monitors, true);
 }
 
-
 void monitor_apply_layout(HSMonitor* monitor) {
     if (monitor) {
         if (*g_monitors_locked) {
@@ -77,6 +81,9 @@ void monitor_apply_layout(HSMonitor* monitor) {
             rect.width -= *g_frame_gap;
         }
         monitor_restack(monitor);
+        if (get_current_monitor() == monitor) {
+            frame_focus_recursive(monitor->tag->frame);
+        }
         if (monitor->tag->floating) {
             frame_apply_floating_layout(monitor->tag->frame, monitor);
         } else {
@@ -85,9 +92,6 @@ void monitor_apply_layout(HSMonitor* monitor) {
                 frame_update_frame_window_visibility(monitor->tag->frame);
             }
         }
-        if (get_current_monitor() == monitor) {
-            frame_focus_recursive(monitor->tag->frame);
-        }
         // remove all enternotify-events from the event queue that were
         // generated while arranging the clients on this monitor
         XEvent ev;
@@ -96,19 +100,49 @@ void monitor_apply_layout(HSMonitor* monitor) {
     }
 }
 
-int list_monitors(int argc, char** argv, GString** output) {
+int list_monitors(int argc, char** argv, GString* output) {
     (void)argc;
     (void)argv;
     int i;
+    GString* monitor_name = g_string_new("");
     for (i = 0; i < g_monitors->len; i++) {
         HSMonitor* monitor = monitor_with_index(i);
-        g_string_append_printf(*output, "%d: %dx%d%+d%+d with tag \"%s\"%s\n",
+        if (monitor->name != NULL ) {
+            g_string_printf(monitor_name, ", named \"%s\"",
+                            monitor->name->str);
+        } else {
+            g_string_truncate(monitor_name, 0);
+        }
+        g_string_append_printf(output, "%d: %dx%d%+d%+d with tag \"%s\"%s%s%s\n",
             i,
             monitor->rect.width, monitor->rect.height,
             monitor->rect.x, monitor->rect.y,
             monitor->tag ? monitor->tag->name->str : "???",
-            (g_cur_monitor == i) ? " [FOCUS]" : "");
+            monitor_name->str,
+            (g_cur_monitor == i) ? " [FOCUS]" : "",
+            monitor->lock_tag ? " [LOCKED]" : "");
     }
+    g_string_free(monitor_name, true);
+    return 0;
+}
+
+int list_padding(int argc, char** argv, GString* output) {
+    HSMonitor* monitor;
+    if (argc < 2) {
+        monitor = get_current_monitor();
+    } else {
+        monitor = string_to_monitor(argv[1]);
+        if (monitor == NULL) {
+            g_string_append_printf(output,
+                "%s: Monitor \"%s\" not found!\n", argv[0], argv[1]);
+            return HERBST_INVALID_ARGUMENT;
+        }
+    }
+    g_string_append_printf(output, "%d %d %d %d\n",
+        monitor->pad_up,
+        monitor->pad_right,
+        monitor->pad_down,
+        monitor->pad_left);
     return 0;
 }
 
@@ -232,11 +266,10 @@ static RectList* disjoin_rects(XRectangle* buf, size_t count) {
 }
 
 
-int disjoin_rects_command(int argc, char** argv, GString** output) {
+int disjoin_rects_command(int argc, char** argv, GString* output) {
     (void)SHIFT(argc, argv);
     if (argc < 1) {
-        g_string_append_printf(*output, "At least one rect is required.\n");
-        return HERBST_INVALID_ARGUMENT;
+        return HERBST_NEED_MORE_ARGS;
     }
     XRectangle* buf = g_new(XRectangle, argc);
     for (int i = 0; i < argc; i++) {
@@ -246,7 +279,7 @@ int disjoin_rects_command(int argc, char** argv, GString** output) {
     RectList* rects = disjoin_rects(buf, argc);
     for (RectList* cur = rects; cur; cur = cur->next) {
         XRectangle r = cur->rect;
-        g_string_append_printf(*output, "%dx%d%+d%+d\n",
+        g_string_append_printf(output, "%dx%d%+d%+d\n",
             r.width, r.height, r.x, r.y);
     }
     rectlist_free(rects);
@@ -254,11 +287,10 @@ int disjoin_rects_command(int argc, char** argv, GString** output) {
     return 0;
 }
 
-int set_monitor_rects_command(int argc, char** argv, GString** output) {
+int set_monitor_rects_command(int argc, char** argv, GString* output) {
     (void)SHIFT(argc, argv);
     if (argc < 1) {
-        g_string_append_printf(*output, "At least one monitor is required.\n");
-        return HERBST_INVALID_ARGUMENT;
+        return HERBST_NEED_MORE_ARGS;
     }
     XRectangle* templates = g_new0(XRectangle, argc);
     for (int i = 0; i < argc; i++) {
@@ -266,6 +298,13 @@ int set_monitor_rects_command(int argc, char** argv, GString** output) {
     }
     int status = set_monitor_rects(templates, argc);
     g_free(templates);
+    if (status == HERBST_TAG_IN_USE) {
+        g_string_append_printf(output,
+            "%s: There are not enough free tags\n", argv[0]);
+    } else if (status == HERBST_INVALID_ARGUMENT) {
+        g_string_append_printf(output,
+            "%s: Need at least one rectangle\n", argv[0]);
+    }
     return status;
 }
 
@@ -279,13 +318,13 @@ int set_monitor_rects(XRectangle* templates, size_t count) {
         HSMonitor* m = monitor_with_index(i);
         m->rect = templates[i];
     }
-    // add aditional monitors
+    // add additional monitors
     for (; i < count; i++) {
         tag = find_unused_tag();
         if (!tag) {
             return HERBST_TAG_IN_USE;
         }
-        add_monitor(templates[i], tag);
+        add_monitor(templates[i], tag, NULL);
         frame_show_recursive(tag->frame);
     }
     // remove monitors if there are too much
@@ -296,11 +335,60 @@ int set_monitor_rects(XRectangle* templates, size_t count) {
     return 0;
 }
 
-HSMonitor* add_monitor(XRectangle rect, HSTag* tag) {
+int find_monitor_index_by_name(char* name) {
+    int i;
+    for (i = 0; i < g_monitors->len; i++) {
+        HSMonitor* mon = monitor_with_index(i);
+        if (mon != NULL && mon->name != NULL && !strcmp(mon->name->str, name)) {
+            return i;
+        }
+    }
+    return -1;
+}
+
+HSMonitor* find_monitor_by_name(char* name) {
+    int i = find_monitor_index_by_name(name);
+    if (i == -1) {
+        return NULL;
+    } else {
+        return monitor_with_index(i);
+    }
+}
+
+int string_to_monitor_index(char* string) {
+    if (string[0] == '\0') {
+        return g_cur_monitor;
+    } else if (string[0] == '-' || string[0] == '+') {
+        // relative monitor index
+        int idx = g_cur_monitor + atoi(string);
+        idx %= g_monitors->len;
+        idx += g_monitors->len;
+        idx %= g_monitors->len;
+        return idx;
+    } else if (isdigit(string[0])) {
+        // absolute monitor index
+        int idx = atoi(string);
+        if (idx < 0 || idx >= g_monitors->len) {
+            return -1;
+        }
+        return idx;
+    } else {
+        // monitor string
+        return find_monitor_index_by_name(string);
+    }
+}
+
+HSMonitor* string_to_monitor(char* string) {
+    int idx = string_to_monitor_index(string);
+    return monitor_with_index(idx);
+}
+
+HSMonitor* add_monitor(XRectangle rect, HSTag* tag, char* name) {
     assert(tag != NULL);
     HSMonitor* m = g_new0(HSMonitor, 1);
     m->rect = rect;
     m->tag = tag;
+    m->name = (name ? g_string_new(name) : NULL);
     m->mouse.x = 0;
     m->mouse.y = 0;
     m->dirty = true;
@@ -312,46 +400,80 @@ HSMonitor* add_monitor(XRectangle rect, HSTag* tag) {
     return g_array_index(g_monitors, HSMonitor*, g_monitors->len-1);
 }
 
-int add_monitor_command(int argc, char** argv) {
-    // usage: add_monitor RECTANGLE TAG [PADUP [PADRIGHT [PADDOWN [PADLEFT]]]]
+int add_monitor_command(int argc, char** argv, GString* output) {
+    // usage: add_monitor RECTANGLE [TAG [NAME]]
     if (argc < 2) {
-        return HERBST_INVALID_ARGUMENT;
+        return HERBST_NEED_MORE_ARGS;
     }
     XRectangle rect = parse_rectangle(argv[1]);
     HSTag* tag = NULL;
+    char* name = NULL;
     if (argc == 2 || !strcmp("", argv[2])) {
         tag = find_unused_tag();
         if (!tag) {
+            g_string_append_printf(output,
+                "%s: There are not enough free tags\n", argv[0]);
             return HERBST_TAG_IN_USE;
         }
     }
     else {
         tag = find_tag(argv[2]);
         if (!tag) {
+            g_string_append_printf(output,
+                "%s: The tag \"%s\" does not exist\n", argv[0], argv[2]);
             return HERBST_INVALID_ARGUMENT;
         }
     }
     if (find_monitor_with_tag(tag)) {
+        g_string_append_printf(output,
+            "%s: The tag \"%s\" is already viewed on a monitor\n", argv[0], argv[2]);
         return HERBST_TAG_IN_USE;
     }
-    HSMonitor* monitor = add_monitor(rect, tag);
-    if (argc > 3 && argv[3][0] != '\0') monitor->pad_up       = atoi(argv[3]);
-    if (argc > 4 && argv[4][0] != '\0') monitor->pad_right    = atoi(argv[4]);
-    if (argc > 5 && argv[5][0] != '\0') monitor->pad_down     = atoi(argv[5]);
-    if (argc > 6 && argv[6][0] != '\0') monitor->pad_left     = atoi(argv[6]);
+    if (argc > 3) {
+        name = argv[3];
+        if (isdigit(name[0])) {
+            g_string_append_printf(output,
+                "%s: The monitor name may not start with a number\n", argv[0]);
+            return HERBST_INVALID_ARGUMENT;
+        }
+        if (!strcmp("", name)) {
+            g_string_append_printf(output,
+                "%s: An empty monitor name is not permitted\n", argv[0]);
+            return HERBST_INVALID_ARGUMENT;
+        }
+        if (find_monitor_by_name(name)) {
+            g_string_append_printf(output,
+                "%s: A monitor with the same name already exists\n", argv[0]);
+            return HERBST_INVALID_ARGUMENT;
+        }
+    }
+    HSMonitor* monitor = add_monitor(rect, tag, name);
     frame_show_recursive(tag->frame);
     monitor_apply_layout(monitor);
     emit_tag_changed(tag, g_monitors->len - 1);
     return 0;
 }
 
-int remove_monitor_command(int argc, char** argv) {
+int remove_monitor_command(int argc, char** argv, GString* output) {
     // usage: remove_monitor INDEX
     if (argc < 2) {
+        return HERBST_NEED_MORE_ARGS;
+    }
+    int index = string_to_monitor_index(argv[1]);
+    if (index == -1) {
+        g_string_append_printf(output,
+            "%s: Monitor \"%s\" not found!\n", argv[0], argv[1]);
         return HERBST_INVALID_ARGUMENT;
     }
-    int index = atoi(argv[1]);
-    return remove_monitor(index);
+    int ret = remove_monitor(index);
+    if (ret == HERBST_INVALID_ARGUMENT) {
+        g_string_append_printf(output,
+            "%s: Index needs to be between 0 and %d\n", argv[0], g_monitors->len - 1);
+    } else if (ret == HERBST_FORBIDDEN) {
+        g_string_append_printf(output,
+            "%s: Can't remove the last monitor\n", argv[0]);
+    }
+    return ret;
 }
 
 int remove_monitor(int index) {
@@ -375,7 +497,7 @@ int remove_monitor(int index) {
     stack_remove_slice(g_monitor_stack, monitor->slice);
     slice_destroy(monitor->slice);
     XDestroyWindow(g_display, monitor->stacking_window);
-    // and remove monitor completly
+    // and remove monitor completely
     g_free(monitor);
     g_array_remove_index(g_monitors, index);
     if (g_cur_monitor >= g_monitors->len) {
@@ -386,22 +508,25 @@ int remove_monitor(int index) {
     return 0;
 }
 
-int move_monitor_command(int argc, char** argv) {
+int move_monitor_command(int argc, char** argv, GString* output) {
     // usage: move_monitor INDEX RECT [PADUP [PADRIGHT [PADDOWN [PADLEFT]]]]
     // moves monitor with number to RECT
     if (argc < 3) {
-        return HERBST_INVALID_ARGUMENT;
+        return HERBST_NEED_MORE_ARGS;
     }
-    int index = atoi(argv[1]);
-    if (index < 0 || index >= g_monitors->len) {
+    HSMonitor* monitor = string_to_monitor(argv[1]);
+    if (monitor == NULL) {
+        g_string_append_printf(output,
+            "%s: Monitor \"%s\" not found!\n", argv[0], argv[1]);
         return HERBST_INVALID_ARGUMENT;
     }
     XRectangle rect = parse_rectangle(argv[2]);
     if (rect.width < WINDOW_MIN_WIDTH || rect.height < WINDOW_MIN_HEIGHT) {
+        g_string_append_printf(output,
+            "%s: Rectangle is too small\n", argv[0]);
         return HERBST_INVALID_ARGUMENT;
     }
     // else: just move it:
-    HSMonitor* monitor = monitor_with_index(index);
     monitor->rect = rect;
     if (argc > 3 && argv[3][0] != '\0') monitor->pad_up       = atoi(argv[3]);
     if (argc > 4 && argv[4][0] != '\0') monitor->pad_right    = atoi(argv[4]);
@@ -411,41 +536,74 @@ int move_monitor_command(int argc, char** argv) {
     return 0;
 }
 
-int monitor_rect_command(int argc, char** argv, GString** result) {
+int rename_monitor_command(int argc, char** argv, GString* output) {
+    if (argc < 3) {
+        return HERBST_NEED_MORE_ARGS;
+    }
+    HSMonitor* mon = string_to_monitor(argv[1]);
+    if (mon == NULL) {
+        g_string_append_printf(output,
+            "%s: Monitor \"%s\" not found!\n", argv[0], argv[1]);
+        return HERBST_INVALID_ARGUMENT;
+    }
+    if (isdigit(argv[2][0])) {
+        g_string_append_printf(output,
+            "%s: The monitor name may not start with a number\n", argv[0]);
+        return HERBST_INVALID_ARGUMENT;
+    } else if (!strcmp("", argv[2])) {
+        // empty name -> clear name
+        if (mon->name != NULL) {
+            g_string_free(mon->name, true);
+            mon->name = NULL;
+        }
+        return 0;
+    }
+    if (find_monitor_by_name(argv[2])) {
+        g_string_append_printf(output,
+            "%s: A monitor with the same name already exists\n", argv[0]);
+        return HERBST_INVALID_ARGUMENT;
+    }
+    if (mon->name == NULL) {
+        // not named before
+        GString* name = g_string_new(argv[2]);
+        mon->name = name;
+    } else {
+        // already named
+        g_string_assign(mon->name, argv[2]);
+    }
+    return 0;
+}
+
+int monitor_rect_command(int argc, char** argv, GString* output) {
     // usage: monitor_rect [[-p] INDEX]
-    char* index_str = NULL;
+    char* monitor_str = NULL;
     HSMonitor* m = NULL;
     bool with_pad = false;
 
-    // if index is supplied
+    // if monitor is supplied
     if (argc > 1) {
-        index_str = argv[1];
+        monitor_str = argv[1];
     }
     // if -p is supplied
     if (argc > 2) {
-        index_str = argv[2];
+        monitor_str = argv[2];
         if (!strcmp("-p", argv[1])) {
             with_pad = true;
         } else {
-            fprintf(stderr, "monitor_rect_command: invalid argument \"%s\"\n",
-                    argv[1]);
+            g_string_append_printf(output,
+                "%s: Invalid argument \"%s\"\n", argv[0], argv[1]);
             return HERBST_INVALID_ARGUMENT;
         }
     }
     // if an index is set
-    if (index_str) {
-        int index;
-        if (1 == sscanf(index_str, "%d", &index)) {
-            m = monitor_with_index(index);
-            if (!m) {
-                fprintf(stderr,"monitor_rect_command: invalid index \"%s\"\n",
-                        index_str);
-                return HERBST_INVALID_ARGUMENT;
-            }
+    if (monitor_str) {
+        m = string_to_monitor(monitor_str);
+        if (m == NULL) {
+            g_string_append_printf(output,
+                "%s: Monitor \"%s\" not found!\n", argv[0], monitor_str);
+            return HERBST_INVALID_ARGUMENT;
         }
-    }
-
-    if (!m) {
+    } else {
         m = get_current_monitor();
     }
     XRectangle rect = m->rect;
@@ -455,20 +613,21 @@ int monitor_rect_command(int argc, char** argv, GString** result) {
         rect.y += m->pad_up;
         rect.height -= m->pad_up + m->pad_down;
     }
-    g_string_printf(*result, "%d %d %d %d",
+    g_string_append_printf(output, "%d %d %d %d",
                     rect.x, rect.y, rect.width, rect.height);
     return 0;
 }
 
-int monitor_set_pad_command(int argc, char** argv) {
+int monitor_set_pad_command(int argc, char** argv, GString* output) {
     if (argc < 2) {
+        return HERBST_NEED_MORE_ARGS;
+    }
+    HSMonitor* monitor = string_to_monitor(argv[1]);
+    if (monitor == NULL) {
+        g_string_append_printf(output,
+            "%s: Monitor \"%s\" not found!\n", argv[0], argv[1]);
         return HERBST_INVALID_ARGUMENT;
     }
-    int index = atoi(argv[1]);
-    if (index < 0 || index >= g_monitors->len) {
-        return HERBST_INVALID_ARGUMENT;
-    }
-    HSMonitor* monitor = monitor_with_index(index);
     if (argc > 2 && argv[2][0] != '\0') monitor->pad_up       = atoi(argv[2]);
     if (argc > 3 && argv[3][0] != '\0') monitor->pad_right    = atoi(argv[3]);
     if (argc > 4 && argv[4][0] != '\0') monitor->pad_down     = atoi(argv[4]);
@@ -501,7 +660,7 @@ void ensure_monitors_are_available() {
     };
     ensure_tags_are_available();
     // add monitor with first tag
-    HSMonitor* m = add_monitor(rect, g_array_index(g_tags, HSTag*, 0));
+    HSMonitor* m = add_monitor(rect, g_array_index(g_tags, HSTag*, 0), NULL);
     g_cur_monitor = 0;
     g_cur_frame = m->tag->frame;
 }
@@ -530,14 +689,30 @@ void all_monitors_apply_layout() {
     }
 }
 
-void monitor_set_tag(HSMonitor* monitor, HSTag* tag) {
+int monitor_set_tag(HSMonitor* monitor, HSTag* tag) {
     HSMonitor* other = find_monitor_with_tag(tag);
     if (monitor == other) {
         // nothing to do
-        return;
+        return 0;
+    }
+    if (monitor->lock_tag) {
+        // If the monitor tag is locked, do not change the tag
+        if (other != NULL) {
+            // but if the tag is already visible, change to the
+            // displaying monitor
+            monitor_focus_by_index(monitor_index_of(other));
+            return 0;
+        }
+        return 1;
     }
     if (other != NULL) {
         if (*g_swap_monitors_to_get_tag) {
+            if (other->lock_tag) {
+                // the monitor we want to steal the tag from is
+                // locked. focus that monitor instead
+                monitor_focus_by_index(monitor_index_of(other));
+                return 0;
+            }
             // swap tags
             other->tag = monitor->tag;
             monitor->tag = tag;
@@ -556,7 +731,7 @@ void monitor_set_tag(HSMonitor* monitor, HSTag* tag) {
             emit_tag_changed(other->tag, monitor_index_of(other));
             emit_tag_changed(tag, g_cur_monitor);
         }
-        return;
+        return 0;
     }
     HSTag* old_tag = monitor->tag;
     // 1. show new tag
@@ -583,23 +758,37 @@ void monitor_set_tag(HSMonitor* monitor, HSTag* tag) {
     while (XCheckMaskEvent(g_display, EnterWindowMask, &ev));
     ewmh_update_current_desktop();
     emit_tag_changed(tag, g_cur_monitor);
+    return 0;
 }
 
-int monitor_set_tag_command(int argc, char** argv) {
+int monitor_set_tag_command(int argc, char** argv, GString* output) {
     if (argc < 2) {
-        return HERBST_INVALID_ARGUMENT;
+        return HERBST_NEED_MORE_ARGS;
     }
     HSMonitor* monitor = get_current_monitor();
     HSTag*  tag = find_tag(argv[1]);
     if (monitor && tag) {
-        monitor_set_tag(get_current_monitor(), tag);
+        int ret = monitor_set_tag(monitor, tag);
+        if (ret != 0) {
+            g_string_append_printf(output,
+                "%s: Could not change tag", argv[0]);
+            if (monitor->lock_tag) {
+                g_string_append_printf(output,
+                    " (monitor %d is locked)", monitor_index_of(monitor));
+            }
+            g_string_append_printf(output, "\n");
+        }
+        return ret;
+    } else {
+        g_string_append_printf(output,
+            "%s: Invalid tag \"%s\"\n", argv[0], argv[1]);
+        return HERBST_INVALID_ARGUMENT;
     }
-    return 0;
 }
 
-int monitor_set_tag_by_index_command(int argc, char** argv) {
+int monitor_set_tag_by_index_command(int argc, char** argv, GString* output) {
     if (argc < 2) {
-        return HERBST_INVALID_ARGUMENT;
+        return HERBST_NEED_MORE_ARGS;
     }
     bool skip_visible = false;
     if (argc >= 3 && !strcmp(argv[2], "--skip-visible")) {
@@ -607,17 +796,28 @@ int monitor_set_tag_by_index_command(int argc, char** argv) {
     }
     HSTag* tag = get_tag_by_index_str(argv[1], skip_visible);
     if (!tag) {
+        g_string_append_printf(output,
+            "%s: Invalid index \"%s\"\n", argv[0], argv[1]);
         return HERBST_INVALID_ARGUMENT;
     }
-    monitor_set_tag(get_current_monitor(), tag);
-    return 0;
+    int ret = monitor_set_tag(get_current_monitor(), tag);
+    if (ret != 0) {
+        g_string_append_printf(output,
+            "%s: Could not change tag (maybe monitor is locked?)\n", argv[0]);
+    }
+    return ret;
 }
 
-int monitor_focus_command(int argc, char** argv) {
+int monitor_focus_command(int argc, char** argv, GString* output) {
     if (argc < 2) {
+        return HERBST_NEED_MORE_ARGS;
+    }
+    int new_selection = string_to_monitor_index(argv[1]);
+    if (new_selection == -1) {
+        g_string_append_printf(output,
+            "%s: Monitor \"%s\" not found!\n", argv[0], argv[1]);
         return HERBST_INVALID_ARGUMENT;
     }
-    int new_selection = atoi(argv[1]);
     // really change selection
     monitor_focus_by_index(new_selection);
     return 0;
@@ -685,6 +885,14 @@ void monitor_focus_by_index(int new_selection) {
         && (monitor->rect.y <= ry) && (ry < monitor->rect.y + monitor->rect.height)) {
         // mouse already is on new monitor
     } else {
+        // If the mouse is located in a gap indicated by
+        // mouse_recenter_gap at the outer border of the monitor,
+        // recenter the mouse.
+        if (min(monitor->mouse.x, abs(monitor->mouse.x - monitor->rect.width)) < *g_mouse_recenter_gap
+            || min(monitor->mouse.y, abs(monitor->mouse.y - monitor->rect.height)) < *g_mouse_recenter_gap) {
+            monitor->mouse.x = monitor->rect.width / 2;
+            monitor->mouse.y = monitor->rect.height / 2;
+        }
         new_x = monitor->rect.x + monitor->mouse.x;
         new_y = monitor->rect.y + monitor->mouse.y;
         XWarpPointer(g_display, None, g_root, 0, 0, 0, 0, new_x, new_y);
@@ -730,6 +938,11 @@ HSMonitor* monitor_with_index(int index) {
 }
 
 int monitors_lock_command(int argc, char** argv) {
+    monitors_lock();
+    return 0;
+}
+
+void monitors_lock() {
     // lock-number must never be negative
     // ensure that lock value is valid
     if (*g_monitors_locked < 0) {
@@ -738,10 +951,14 @@ int monitors_lock_command(int argc, char** argv) {
     // increase lock => it is definitely > 0, i.e. locked
     (*g_monitors_locked)++;
     monitors_lock_changed();
-    return 0;
 }
 
 int monitors_unlock_command(int argc, char** argv) {
+    monitors_unlock();
+    return 0;
+}
+
+void monitors_unlock() {
     // lock-number must never be lower than 1 if unlocking
     // so: ensure that lock value is valid
     if (*g_monitors_locked < 1) {
@@ -750,7 +967,6 @@ int monitors_unlock_command(int argc, char** argv) {
     // decrease lock => unlock
     (*g_monitors_locked)--;
     monitors_lock_changed();
-    return 0;
 }
 
 void monitors_lock_changed() {
@@ -769,6 +985,42 @@ void monitors_lock_changed() {
     }
 }
 
+int monitor_lock_tag_command(int argc, char** argv, GString* output) {
+    char* cmd_name = argv[0];
+    (void)SHIFT(argc, argv);
+    HSMonitor *monitor;
+    if (argc >= 1) {
+        monitor = string_to_monitor(argv[0]);
+        if (monitor == NULL) {
+            g_string_append_printf(output,
+                "%s: Monitor \"%s\" not found!\n", cmd_name, argv[0]);
+            return HERBST_INVALID_ARGUMENT;
+        }
+    } else {
+        monitor = get_current_monitor();
+    }
+    monitor->lock_tag = true;
+    return 0;
+}
+
+int monitor_unlock_tag_command(int argc, char** argv, GString* output) {
+    char* cmd_name = argv[0];
+    (void)SHIFT(argc, argv);
+    HSMonitor *monitor;
+    if (argc >= 1) {
+        monitor = string_to_monitor(argv[0]);
+        if (monitor == NULL) {
+            g_string_append_printf(output,
+                "%s: Monitor \"%s\" not found!\n", cmd_name, argv[0]);
+            return HERBST_INVALID_ARGUMENT;
+        }
+    } else {
+        monitor = get_current_monitor();
+    }
+    monitor->lock_tag = false;
+    return 0;
+}
+
 // monitor detection using xinerama (if available)
 #ifdef XINERAMA
 // inspired by dwm's isuniquegeom()
@@ -821,7 +1073,7 @@ bool detect_monitors_xinerama(XRectangle** ret_rects, size_t* ret_count) {
 
 #endif /* XINERAMA */
 
-// monitor detaction that always works: one monitor across the entire screen
+// monitor detection that always works: one monitor across the entire screen
 bool detect_monitors_simple(XRectangle** ret_rects, size_t* ret_count) {
     *ret_count = 1;
     *ret_rects = g_new0(XRectangle, 1);
@@ -832,7 +1084,7 @@ bool detect_monitors_simple(XRectangle** ret_rects, size_t* ret_count) {
     return true;
 }
 
-int detect_monitors_command(int argc, char **argv) {
+int detect_monitors_command(int argc, char **argv, GString* output) {
     MonitorDetection detect[] = {
         detect_monitors_xinerama,
         detect_monitors_simple,
@@ -851,41 +1103,49 @@ int detect_monitors_command(int argc, char **argv) {
     // apply it
     int ret = set_monitor_rects(monitors, count);
     g_free(monitors);
+    if (ret == HERBST_TAG_IN_USE && output != NULL) {
+        g_string_append_printf(output,
+            "%s: There are not enough free tags\n", argv[0]);
+    }
     return ret;
 }
 
-int monitor_stack_window_count() {
-    return stack_window_count(g_monitor_stack);
+int monitor_stack_window_count(bool only_clients) {
+    return stack_window_count(g_monitor_stack, only_clients);
 }
 
-void monitor_stack_to_window_buf(Window* buf, int len, int* remain_len) {
-    stack_to_window_buf(g_monitor_stack, buf, len, remain_len);
+void monitor_stack_to_window_buf(Window* buf, int len, bool only_clients,
+                                 int* remain_len) {
+    stack_to_window_buf(g_monitor_stack, buf, len, only_clients, remain_len);
 }
 
 HSStack* get_monitor_stack() {
     return g_monitor_stack;
 }
 
-int monitor_raise_command(int argc, char** argv) {
+int monitor_raise_command(int argc, char** argv, GString* output) {
+    char* cmd_name = argv[0];
     (void)SHIFT(argc, argv);
     HSMonitor* monitor;
     if (argc >= 1) {
-        monitor = monitor_with_index(atoi(argv[0]));
+        monitor = string_to_monitor(argv[0]);
+        if (monitor == NULL) {
+            g_string_append_printf(output,
+                "%s: Monitor \"%s\" not found!\n", cmd_name, argv[0]);
+            return HERBST_INVALID_ARGUMENT;
+        }
     } else {
         monitor = get_current_monitor();
     }
-    if (!monitor) {
-        return HERBST_INVALID_ARGUMENT;
-    }
     stack_raise_slide(g_monitor_stack, monitor->slice);
     return 0;
 }
 
 void monitor_restack(HSMonitor* monitor) {
-    int count = 1 + stack_window_count(monitor->tag->stack);
+    int count = 1 + stack_window_count(monitor->tag->stack, false);
     Window* buf = g_new(Window, count);
     buf[0] = monitor->stacking_window;
-    stack_to_window_buf(monitor->tag->stack, buf + 1, count - 1, NULL);
+    stack_to_window_buf(monitor->tag->stack, buf + 1, count - 1, false, NULL);
     /* remove a focused fullscreen client */
     Window win = frame_focused_window(monitor->tag->frame);
     HSClient* client = win ? get_client_from_window(win) : NULL;
@@ -900,3 +1160,18 @@ void monitor_restack(HSMonitor* monitor) {
     g_free(buf);
 }
 
+int shift_to_monitor(int argc, char** argv, GString* output) {
+    if (argc <= 1) {
+        return HERBST_NEED_MORE_ARGS;
+    }
+    char* monitor_str = argv[1];
+    HSMonitor* monitor = string_to_monitor(monitor_str);
+    if (!monitor) {
+        g_string_append_printf(output,
+            "%s: Invalid monitor\n", monitor_str);
+        return HERBST_INVALID_ARGUMENT;
+    }
+    tag_move_focused_client(monitor->tag);
+    return 0;
+}
+
