@@ -55,9 +55,10 @@ typedef struct {
     GString*        tag_name;
     GString*        tree_index;
     bool            focus; // if client should get focus
-    bool            switchtag; // if the tag may be switchted for focusing it
-    bool            manage; // wether we should manage it
+    bool            switchtag; // if the tag may be switched for focusing it
+    bool            manage; // whether we should manage it
     bool            fullscreen;
+    bool            ewmhnotify; // whether to send ewmh-notifications
 } HSClientChanges;
 
 void rules_init();
@@ -70,9 +71,8 @@ void client_changes_free_members(HSClientChanges* changes);
 HSRule* rule_create();
 void rule_destroy(HSRule* rule);
 
-int rule_add_command(int argc, char** argv);
-int rule_remove_command(int argc, char** argv);
-
+int rule_add_command(int argc, char** argv, GString* output);
+int rule_remove_command(int argc, char** argv, GString* output);
 
 #endif
 
